// [20/08/23] Created
// Low-ohm Meter (AVR Version) FW
// AVR Clock: 2MHz
// Fuse bits: -fl11100010 -fh1101100
// (C) 2020 Hiroyuki Yamada

#include "main.h"
#include "adc.h"

#define FWREV 0x0100

volatile uint8_t g_segmentData[3];
const uint8_t g_fontData[] = // 0-9, a-f
{
	0b00111111, 0b00000110, 0b01011011, 0b01001111, 0b01100110, 
	0b01101101, 0b01111101, 0b00000111, 0b01111111, 0b01101111,
	0b01110111, 0b01111100, 0b00111001, 0b01011110, 0b01111001, 0b01110001
};


static void SREGSend(uint8_t a) // LED Shift Register
{
	uint8_t j;
	for(j=0; j<8; j++){
		if(a & 0x80) SBI(PORTB, PB3);
		else CBI(PORTB, PB3);
		// Clock
		SBI(PORTB, PB5);
		CBI(PORTB, PB5);
		a = a << 1;
	}
	// Latch Pulse
	SBI(PORTB, PB0);
	CBI(PORTB, PB0);
}

volatile uint8_t  g_100msFlag;
// timer0 : 1ms
ISR(TIMER0_OVF_vect)
{
	static uint8_t phase, localCnt;

	PORTD |= 0xE0;
	
	phase++;
	if(phase > 4) phase = 0;
	if(phase < 3){
		SREGSend(~g_segmentData[phase]);
		CBI(PORTD, (5+phase));
	}else{
	}
	
	localCnt ++;
	if(localCnt > 200){
		localCnt = 0;
		g_100msFlag = 1;
	}
	
	TCNT0 = 128;
}


ISR(TIMER2_OVF_vect)
{
	// Activate ADC
	ADCSRA |= 0x40; // start conversion

	TCCR2  = 0; // Timer2 OFF
	// debug
	SBI(PORTC, PC3);
}

//#define TIMER2_REF_CNT 1042	// CKSEL = 1MHz
#define TIMER2_REF_CNT 2084	// CKSEL = 2MHz

// OC1A 1 -> 0 (Output H)
ISR(TIMER1_COMPA_vect)
{
//	CBI(PORTB, PB2);
	static uint8_t g_adcPhase;
	if((g_adcPhase++ & 0xFF) == 0) SetADCChannel(0);
	TCNT2  = -TIMER2_REF_CNT/4/4;
	TCCR2  = 0b00000011; // 1M/8/32
}
// OVF (Output L)
ISR(TIMER1_CAPT_vect)
{
	TCNT2  = -TIMER2_REF_CNT/4/4;
	TCCR2  = 0b00000011; // 1M/8/32
}


static void InitPort(void)
{
	
	// init clock
//	CLKPR = 0x80;
//	CLKPR = 0x00; // CLKPS = 0 (1MHz)
	
//	MCUSR = 0;
	
	// init pins
	DDRB  = 0b00101111; // PB0,3,5: SREG, PB1: OC1A, PB2: OC1B
	PORTB = 0b11111111; // 
	
	DDRC  = 0b00111000; // PC0-1: ADC IN, PC4-5: GAIN SEL
	PORTC = 0b11000100;
	
	DDRD  = 0b11111111; // PD0-1: LED, PD4: POWER, PD5-7: COM
	PORTD = 0b11111100; // Power Hold
	
}

// Initialize
static void Init(void)
{
	
	InitPort();
	_delay_us(100);
	
	// init timer
	// timer0 (2M/8/256 = 4ms)
	TCCR0 = 0b00000010; // CLK/8
	TIMSK = 0b00000001; // TOIE0
	
	// timer1 OC1A and OC1B output (2M/8/2048 = 120Hz)
	TCCR1A = 0b11100010;
	TCCR1B = 0b00011010; // High-speed PWM, TOP = ICR1
	ICR1   = TIMER2_REF_CNT;
	OCR1A  = TIMER2_REF_CNT/2;
	OCR1B  = TIMER2_REF_CNT/2;
	TIMSK |= 0b00110000; // TICIE1 = 1, OCIE1A = 1, OCIE1B = 0
	
	// timer2 
	TIMSK |= 0b01000000; // TOIE2 = 1
	
	// init ADC
	ADC_Init();
	
	sei();
	
}

int main(void)
{
	uint16_t ovfCounter = 0, powerUpConter = 0;
	uint8_t i;
	for(i=0; i<3; i++) g_segmentData[i] = g_fontData[i];
	uint8_t curDisplayRange = 0;
	
	Init();
	SBI(PORTD, PD0);
	
	
	while(1)
	{
		if( g_100msFlag && (GetBufferCount(0) == ADC_BUFFER_SIZE) )
		{
			powerUpConter ++;
			uint16_t adcVal, vccVal;
			float rx = 0, vcc = 0;
			
			g_100msFlag = 0;
			vccVal = GetBufferAvg(2, 1) - GetBufferAvg(3, 1);
			
			uint16_t v0 = GetBufferAvg(0, 1);
			uint16_t v1 = GetBufferAvg(1, 1);
			vcc = vccVal * VREF_VOLTAGE / 1024 / ADC_BUFFER_SIZE;
			adcVal = (v0>v1)? v0-v1: 0; // Get sum of buffer
			rx = adcVal / (vccVal * 2 * GetCurrentGain() - adcVal) * RSRC / 2;
			
			// When overflow continues with 10 min (6000 counts)
			if(powerUpConter > 6000)
			{
				// Auto power off
				CBI(PORTD, 4); // /DCDC_SHDN
			}
			
			if(rx > OVERRANGE_THRESHOLD) // Over Range
			{
				g_segmentData[0] = 0;
				g_segmentData[1] = 0x80 | g_fontData[0]; // O.
				g_segmentData[2] = 0x80 | g_fontData[0x0F]; // F.
				ovfCounter ++;
				
				CBI(PORTD, PD1);
				CBI(PORTD, PD0);
				
				// When overflow continues with 1 min (600 counts)
				if(ovfCounter > 600)
				{
					// Auto power off
					CBI(PORTD, 4); // /DCDC_SHDN
				}
			}
			else
			{
				if(rx < 0.0995) curDisplayRange = 0;
				else if(rx > 0.0999 && rx < 0.995) curDisplayRange = 1;
				else if(rx > 0.999 && rx < 9.95) curDisplayRange = 2;
				else if(rx > 9.99) curDisplayRange = 3;
				
				ovfCounter = 0;
				if(curDisplayRange == 0)
				{
					uint8_t d1, d2;
					adcVal = rx * 10000;// xx.x m ohm
					if(adcVal >= 999) adcVal = 999;
				
					// show adc value
					d1 = (adcVal / 100)% 10;
					g_segmentData[0] = d1? g_fontData[d1]: 0;
					d2 = (adcVal / 10)% 10;
					g_segmentData[1] = g_fontData[d2] | 0x80;
					g_segmentData[2] = g_fontData[(adcVal )% 10];
					g_fontData[(adcVal )% 10] | 0x80;
					
					CBI(PORTD, PD0);
					SBI(PORTD, PD1);
				}
				else if(curDisplayRange == 1)
				{
					uint8_t d1, d2;
					adcVal = rx * 1000;// xxx m ohm
					if(adcVal >= 999) adcVal = 999;
				
					// show adc value
					d1 = (adcVal / 100)% 10;
					g_segmentData[0] = d1? g_fontData[d1]: 0;
					d2 = (adcVal / 10)% 10;
					g_segmentData[1] = (d1 || d2)? g_fontData[d2]: 0;
					g_segmentData[2] = g_fontData[(adcVal )% 10] | 0x80;
					
					CBI(PORTD, PD0);
					SBI(PORTD, PD1);
				}
				else if(curDisplayRange == 2)
				{
					adcVal = rx * 100;// x.xx ohm
				
					// show adc value
					g_segmentData[0] = g_fontData[(adcVal / 100)% 10] | 0x80;
					g_segmentData[1] = g_fontData[(adcVal / 10)% 10];
					g_segmentData[2] = g_fontData[(adcVal )% 10];
					g_fontData[(adcVal )% 10] | 0x80;
					
					SBI(PORTD, PD0);
					CBI(PORTD, PD1);
				}
				else
				{
					adcVal = rx * 10;// xx.x ohm
				
					// show adc value
					g_segmentData[0] = g_fontData[(adcVal / 100)% 10];
					g_segmentData[1] = g_fontData[(adcVal / 10)% 10] | 0x80;
					g_segmentData[2] = g_fontData[(adcVal )% 10];
					g_fontData[(adcVal )% 10] | 0x80;
					
					SBI(PORTD, PD0);
					CBI(PORTD, PD1);
				}
			}
		
		}
	}
}
